from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.decorators import permission_classes
from rest_framework.permissions import AllowAny
from rest_framework.response import Response
from rest_framework.views import APIView
from scoped_rbac.rest import AccessControlledAPIView

from scrud_django import collection_type_uri_for, services
from scrud_django.decorators import scrudful_api_view, scrudful_viewset
from scrud_django.mixins.view_mixins import ResourceViewMixin
from scrud_django.models import Resource
from scrud_django.registration import get_resource_type_for
from scrud_django.validators import JsonValidator


@scrudful_viewset
class ResourceViewSet(
    AccessControlledAPIView, ResourceViewMixin, viewsets.ModelViewSet
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_type_name = kwargs.get('resource_type_name')
        self.validator = JsonValidator()

    @property
    def resource_type_iri(self):
        resource_type = self.get_resource_type()
        return resource_type.type_uri

    @property
    def list_type_iri(self):
        return collection_type_uri_for(self.resource_type_iri)


@scrudful_viewset
class ResourceViewSetGenericFilter(
    AccessControlledAPIView, ResourceViewMixin, viewsets.ModelViewSet
):
    filter_backends = [DjangoFilterBackend]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resource_type_name = kwargs.get('resource_type_name')
        self.validator = JsonValidator()

    def list(self, request):
        """Return the resource for the given resource type name."""
        resource_type = self.get_resource_type()

        query_filter = dict(request.GET.items())

        queryset = Resource.objects.filter(resource_type=resource_type).filter(
            content__contains=query_filter
        )

        queryset = self.filter_queryset(queryset)

        page = self.paginate_queryset(queryset)

        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return Response(self.get_paginated_response(serializer.data))

        serializer = self.get_serializer(instance=queryset, many=True)
        return Response(serializer.data)

    @property
    def resource_type_iri(self):
        resource_type = self.get_resource_type()
        return resource_type.type_uri

    @property
    def list_type_iri(self):
        return collection_type_uri_for(self.resource_type_iri)


class ResourceCollectionSchemaView(APIView):
    permission_classes = [AllowAny]

    def get(self, request, slug: str):
        resource_type = get_resource_type_for(slug)

        if resource_type.schema:
            content_defn = resource_type.schema.content
        elif resource_type.schema_uri:
            content_defn = {"$ref": resource_type.schema_uri}
        else:
            content_defn = {"type": "any"}

        schema = {
            "$id": "https://api.openteams.com/json-schema/ResourceCollection"
            f"?contents_type={resource_type.type_uri}",
            "$schema": "http://json-schema.org/draft-04/schema",
            "description": f"A listing of resources of type {resource_type.type_uri}.",
            "properties": {
                "count": {
                    "type": "integer",
                    "description": "The total number of items in the collection.",
                },
                "page_count": {
                    "type": "integer",
                    "description": "The total number of pages in the collection.",
                },
                "first": {
                    "type": "string",
                    "format": "uri",
                    "description": "URL of the first page.",
                },
                "previous": {
                    "type": "string",
                    "format": "uri",
                    "description": "URL of the previous page, if any.",
                },
                "next": {
                    "type": "string",
                    "format": "uri",
                    "description": "URL of the next page, if any.",
                },
                "last": {
                    "type": "string",
                    "format": "uri",
                    "description": "URL of the last page.",
                },
                "content": {
                    "properties": {
                        "type": "array",
                        "description": f"Listing of {resource_type.type_uri} "
                        "resources in Envelopes.",
                        "items": {
                            "properties": {
                                "href": {
                                    "type": "string",
                                    "format": "uri",
                                    "description": "URL of the nested resource.",
                                },
                                "etag": {
                                    "type": "string",
                                    "description": "HTTP ETag header of the nested "
                                    "resource.",
                                },
                                "last_modified": {
                                    "type": "string",
                                    "description": "HTTP Last-Modified header of the "
                                    "nested resource.",
                                },
                                "content": content_defn,
                            },
                        },
                    },
                },
            },
        }

        return Response(schema)


class ResourceCollectionContextView(APIView):
    permission_classes = [AllowAny]

    def get(self, request, slug: str):
        resource_type = get_resource_type_for(slug)
        return Response(
            {
                "openteams": "https://api.openteams.com/json-ld/",
                "@id": collection_type_uri_for(resource_type.type_uri),
                "http://www.w3.org/2000/01/rdf-schema#subClassOf": {
                    "@id": "https://api.openteams.com/json-ld/ResourceCollection"
                },
                "count": {"@id": "openteams:count"},
                "page_count": {"@id": "openteams:page_count"},
                "first": {"@id": "opententeams:first"},
                "previous": {"@id": "opententeams:previous"},
                "next": {"@id": "opententeams:next"},
                "last": {"@id": "opententeams:last"},
                "content": {
                    "@id": "openteams:Envelope",
                    "@container": "@list",
                    "openteams:envelopeContents": resource_type.type_uri,
                },
            }
        )


@scrudful_api_view(
    etag_func=lambda *args, **kwargs: services.etag,
    last_modified_func=lambda *args, **kwargs: services.last_modified,
)
@permission_classes([AllowAny])
def get_service_list(request, *args, **kwargs):
    catalog = {}
    for type_uri, url_func in services.services.items():
        catalog[type_uri] = url_func(request=request)
    return Response(catalog)



# Generated by Django 2.2.22 on 2021-05-16 17:02

from django.db import migrations, models
from django_jsonfield_backport.models import JSONField


class Migration(migrations.Migration):

    dependencies = [
        ('scrud_django', '0004_auto_20210306_1600'),
    ]

    operations = [
        migrations.AlterField('Resource', 'content', models.fields.TextField()),
        migrations.AlterField('Resource', 'content', JSONField()),
    ]


import json
from datetime import timezone

from django.utils.http import http_date, quote_etag
from rest_framework.reverse import reverse

from scrud_django.headers import get_link_header_for


class SerializerMixin:
    def __init__(self, *args, **kwargs):
        self.resource_type = kwargs.pop("resource_type")
        super().__init__(*args, **kwargs)

    def to_representation(self, instance, envelope=False, context=None):
        if type(instance.content) is str:
            content = json.loads(instance.content)
        else:
            content = instance.content
        if not envelope:
            return content
        request = self._context["request"]
        last_modified = http_date(
            instance.modified_at.replace(tzinfo=timezone.utc).timestamp()
        )
        return {
            'href': reverse(
                self.resource_type.route_name_detail(),
                args=[instance.id],
                request=request,
            ),
            'last_modified': last_modified,
            'etag': quote_etag(instance.etag),
            'link': get_link_header_for(self.resource_type, request),
            'content': content,
        }

    def to_internal_value(self, data):
        return data

    def create(self, validated_data):
        """ Create and return instance """
        return NotImplementedError()

    def update(self, instance, validated_data):
        """ Update and return instance """
        return NotImplementedError()



from rest_framework.reverse import reverse

from scrud_django.utils import link_content


def get_schema_uri_for(resource_type, request):
    uri = None
    # prefer a local version to a remote version
    if resource_type.schema:
        schema = resource_type.schema
        uri = reverse(
            schema.resource_type.route_name_detail(), args=[schema.id], request=request,
        )
    elif resource_type.schema_uri:
        uri = resource_type.schema_uri
    return uri


def get_context_uri_for(resource_type, request):
    uri = None
    # prefer a local version to a remote version
    if resource_type.context:
        context = resource_type.context
        uri = reverse(
            context.resource_type.route_name_detail(),
            args=[context.id],
            request=request,
        )
    elif resource_type.context_uri:
        uri = resource_type.context_uri
    return uri


def get_link_header_for(resource_type, request):
    schema_uri = get_schema_uri_for(resource_type, request)
    context_uri = get_context_uri_for(resource_type, request)
    schema_link = (
        link_content(schema_uri, "describedBy", "application/json")
        if schema_uri
        else ""
    )
    context_link = (
        link_content(
            context_uri, "http://www.w3.org/ns/json-ld#context", "application/ld+json",
        )
        if context_uri
        else ""
    )
    join_links = ", " if schema_link and context_link else ""
    return schema_link + join_links + context_link



from django.db.models import Manager
from rest_framework import serializers

from scrud_django import models
from scrud_django.mixins.serializer_mixins import SerializerMixin
from scrud_django.models import Resource
from scrud_django.registration import ResourceRegistration
from scrud_django.validators import JsonValidator, WorkflowValidator


class EnvelopeSerializer(serializers.ListSerializer):
    def to_representation(self, data):
        iterable = data.all() if isinstance(data, Manager) else data
        return [self.child.to_representation(item, envelope=True) for item in iterable]


class ResourceSerializer(SerializerMixin, serializers.Serializer):
    class Meta:
        model = models.Resource
        list_serializer_class = EnvelopeSerializer
        validators = [JsonValidator(), WorkflowValidator()]

    def create(self, validated_data):
        instance = Resource(content=validated_data, resource_type=self.resource_type)
        return ResourceRegistration.register_instance(instance)

    def update(self, instance, validated_data):
        return ResourceRegistration.update(
            content=validated_data,
            register_type=self.resource_type.slug,
            slug=instance.slug,
        )


class JSONSchemaSerializer(serializers.Serializer):
    """Serializer for JSON-Schema data."""

    class Meta:
        fields = ["$id", "$schema", "title", "description", "properties"]


class JSONLDSerializer(serializers.Serializer):
    """Serializer for JSON-LD data."""

    class Meta:
        fields = ["$id", "$schema", "title", "description", "properties"]



from dataclasses import dataclass
from scrud_django.exceptions import WorkflowTransitionError


@dataclass
class TransitionCheck:
    transition: str


def authorized_transitions(resource, configuration):
    """This permissions Operator function always returns `True`. It's a placeholder
    to enable the `WorkflowValidator` to instead look to see if there is a registered
    policy that dictates the authorized transitions. Otherwise, the workflow
    validation would run twice.
    
    To use this `scoped_rbac.policy.Operator`, define an operator condition that
    specifies the `allowed_transitions` property. Here's an example: ```json {
    "condition": "operator", "allowed_transitions": ["start"] }
    ```

    The name for the initial transition is `__INITIALIZE__`
    """
    if isinstance(resource, TransitionCheck):
        if resource.transition in configuration["authorized_transitions"]:
            return True
        else:
          return False
    return True



from typing import Optional

import reversion
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.response import Response
from rest_framework.reverse import reverse

from scrud_django.headers import get_context_uri_for, get_schema_uri_for
from scrud_django.models import Resource
from scrud_django.paginations import StandardResultsSetPagination
from scrud_django.registration import (
    ResourceRegistration,
    get_resource_type_for,
)
from scrud_django.serializers import ResourceSerializer


class ResourceViewMixin:
    queryset = Resource.objects.all()
    serializer_class = ResourceSerializer
    pagination_class = StandardResultsSetPagination

    # scrud variable
    resource_type_name: Optional[str] = None

    class Meta:
        def etag_func(view_instance, request, slug: str):
            instance = view_instance.get_instance(request, slug)
            return instance.etag

        def last_modified_func(view_instance, request, slug: str):
            instance = view_instance.get_instance(request, slug)
            return instance.modified_at

        def schema_link_or_func(view_instance, request, slug: str = None):
            if view_instance.resource_type_name:
                resource_type = view_instance.get_resource_type()
                return get_schema_uri_for(resource_type, request)
            return None

        def context_link_or_func(view_instance, request, slug: str = None):
            if view_instance.resource_type_name:
                resource_type = view_instance.get_resource_type()
                return get_context_uri_for(resource_type, request)
            return None

        def list_etag_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return resource_type.etag

        def list_last_modified_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return resource_type.modified_at

        def list_schema_link_or_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return reverse(
                "collections-json-schema", args=[resource_type.slug], request=request,
            )

        def list_context_link_or_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return reverse(
                "collections-json-ld", args=[resource_type.slug], request=request,
            )

    def get_resource_type(self):
        return get_resource_type_for(self.resource_type_name)

    def get_instance(self, request, slug: str):
        resource_type = self.get_resource_type()
        instance = get_object_or_404(
            Resource, resource_type=resource_type, pk=int(slug)
        )
        return instance

    def get_serializer(self, *args, **kwargs):
        kwargs["resource_type"] = self.get_resource_type()
        return super().get_serializer(*args, **kwargs)

    def perform_create(self, serializer):
        if self.get_resource_type().uses_reversion:
            with reversion.create_revision():
                response = serializer.save()
                reversion.set_comment('Add Resource')
            return response
        return serializer.save()

    def create(self, request):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = {
            'Location': reverse(
                serializer.instance.resource_type.route_name_detail(),
                args=[serializer.instance.id],
                request=request,
            )
        }
        return Response(
            serializer.data, headers=headers, status=status.HTTP_201_CREATED
        )

    def update(self, request, slug: str):
        """Update a Resource."""
        instance = self.get_instance(request, slug)
        serializer = self.get_serializer(
            data=request.data, instance=instance, many=False
        )
        serializer.is_valid(raise_exception=True)
        instance = ResourceRegistration.update(
            content=request.data, register_type=self.resource_type_name, slug=slug,
        )
        serializer = self.get_serializer(instance=instance, many=False)
        return Response(serializer.data)

    def destroy(self, request, slug: str):
        """Update a Resource."""
        ResourceRegistration.delete(
            register_type=self.resource_type_name, slug=slug,
        )
        return Response(status=status.HTTP_204_NO_CONTENT)

    def retrieve(self, request, slug: str):
        """Return the resource for the given resource type name."""
        instance = self.get_instance(request, slug)
        serializer = self.get_serializer(instance=instance, many=False)
        return Response(serializer.data)

    def list(self, request):
        """Return the resource for the given resource type name."""
        resource_type = self.get_resource_type()

        queryset = Resource.objects.filter(resource_type=resource_type)
        page = self.paginate_queryset(queryset)

        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return Response(self.get_paginated_response(serializer.data))

        serializer = self.get_serializer(instance=queryset, many=True)
        return Response(serializer.data)


class GenericViewMixin:

    # scrud variable
    resource_type_name: Optional[str] = None

    class Meta:
        def etag_func(view_instance, request, slug: str):
            instance = view_instance.get_instance(request, slug)
            return instance.etag

        def last_modified_func(view_instance, request, slug: str):
            instance = view_instance.get_instance(request, slug)
            return instance.modified_at

        def schema_link_or_func(view_instance, request, slug: str = None):
            if view_instance.resource_type_name:
                resource_type = view_instance.get_resource_type()
                return get_schema_uri_for(resource_type, request)
            return None

        def context_link_or_func(view_instance, request, slug: str = None):
            if view_instance.resource_type_name:
                resource_type = view_instance.get_resource_type()
                return get_context_uri_for(resource_type, request)
            return None

        def list_etag_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return resource_type.etag

        def list_last_modified_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return resource_type.modified_at

        def list_schema_link_or_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return reverse(
                "collections-json-schema", args=[resource_type.slug], request=request,
            )

        def list_context_link_or_func(view_instance, request, *args, **kwargs):
            resource_type = view_instance.get_resource_type()
            return reverse(
                "collections-json-ld", args=[resource_type.slug], request=request,
            )

    def get_resource_type(self):
        return get_resource_type_for(self.resource_type_name)

    def get_instance(self, request, slug: str):
        """
            Subclasses **MUST** override this method.
            Return the instance from the DB, used when updating,
            retrieving and deleting the instance.
        """
        raise NotImplementedError()

    def get_serializer(self, *args, **kwargs):
        kwargs["resource_type"] = self.get_resource_type()
        return super().get_serializer(*args, **kwargs)

    def create(self, request):
        """ Add a Resource."""
        raise NotImplementedError()

    def update(self, request, slug: str):
        """Update a Resource."""
        raise NotImplementedError()

    def destroy(self, request, slug: str):
        """Update a Resource."""
        raise NotImplementedError()

    def retrieve(self, request, slug: str):
        """Return the resource for the given resource type name and slug."""
        raise NotImplementedError()

    def list(self, request):
        """Return the list of resources for the given resource type name."""
        raise NotImplementedError()



# Generated by Django 2.2.11 on 2021-05-18 22:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scrud_django', '0005_fix_resource_content_column_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='resourcetype',
            name='uses_reversion',
            field=models.BooleanField(default=True),
        ),
    ]



from django.conf import settings
from rest_framework.settings import import_from_string


DEFAULTS = {
    "DEFAULT_WORKFLOW_POLICIES": {}
}


def workflow_policy_settings():
    return getattr(settings, "SCRUD_WORKFLOW", dict())


def import_policy_or_func(setting):
    default = DEFAULTS[setting]
    policy_or_func = workflow_policy_settings().get(setting, default)
    if isinstance(policy_or_func, str):
        policy_or_func = import_from_string(policy_or_func, setting)
        if callable(policy_or_func):
            policy_or_func = policy_or_func()
    return policy_or_func


def default_workflow_policies():
    return import_policy_or_func("DEFAULT_WORKFLOW_POLICIES")


from datetime import timezone

from django.utils.http import http_date
from haystack import indexes
from jsonpath_ng import parse

from .models import Resource


class ResourceIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=False)
    type_uri = indexes.CharField(use_template=False)
    modified_at = indexes.CharField(use_template=False)
    etag = indexes.CharField(use_template=False)

    def get_model(self):
        return Resource

    def prepare(self, obj):
        self.prepared_data = super().prepare(obj)
        self.prepared_data["type_uri"] = obj.resource_type.type_uri
        self.prepared_data["modified_at"] = http_date(
            obj.modified_at.replace(tzinfo=timezone.utc).timestamp()
        )
        self.prepared_data["etag"] = obj.etag
        indexing_policy = self.get_indexing_policy(obj)
        text_policy = indexing_policy.get("text", None)
        # TODO in order to have an indexing policy per ResourceType in Haystack we'd
        # need to have a Model per ResourceType. We *can* figure that out, but, let's
        # figure out one thing at a time. For now, we'll index only the primary document
        # content using the indexing policy, leaving indexing of fields to a later date.
        # self.prepared_data.update({
        #     field: self.apply(policy, obj.content) for field, policy in
        #         indexing_policy.items()
        # })
        if text_policy:
            self.prepared_data["text"] = self.apply(text_policy, obj.content)
        else:
            self.prepared_data["text"] = ""
        return self.prepared_data

    def get_indexing_policy(self, obj):
        if obj.resource_type.indexing_policy is None:
            return {}
        return obj.resource_type.indexing_policy.content

    def apply(self, policy, content):
        ret = None
        if isinstance(policy, list):
            for p in policy:
                ret = self.concat_result(ret, self.apply(p, content))
        else:
            compiled_policy = parse(policy)
            for match in compiled_policy.find(content):
                ret = self.concat_result(ret, match.value)
        return ret

    def concat_result(self, original, addition):
        if original is None:
            return addition
        if original is not None and addition is None:
            return original
        return original + "\n" + addition
